# -*- coding: utf-8 -*-
"""
Created on Tue Jul 12 09:07:40 2022

@author: janpe
"""

from .fancy_classes import Graph
from .saver import saver
from .lossfunctions import loss_dic
import numpy as np
from scipy import optimize

import logging
log = logging.getLogger(__name__)

class topological_opti:

    def __init__(self, start_graph: Graph, saver: saver, ent_dic=None, target_state=None,
                 config=None, safe_history=True):

        self.config = config
        self.imaginary = self.config['imaginary']
        if self.config['loss_func'] == 'ent':
            self.ent_dic = ent_dic
        else:
            self.target = target_state  # object of State class

        # do preoptimization on complete starting graph, this might already take some time
        self.graph = self.pre_optimize_start_graph(start_graph)
        self.saver = saver
        self.save_hist = safe_history
        self.history = []

    def check(self, result: object, lossfunctions: object):
        """
        check if all loss functions fulfill conditions for success. mostly defined through thresholds.

        Parameters
        ----------
        result : object
             from scipy.mimnimizer class
        lossfunctions : object
            list of all loss functions

        Returns
        -------
        bool
            False if we keep Graph or True if we can delete edge

        """

        if self.config['loss_func'] == 'ent':
            if abs(result.fun) - abs(self.loss_val[0]) > self.config['thresholds'][0]:
                return False
        else:
            # uncomment to see where checks fail
            # print(result.fun, self.config['thresholds'][0])
            if result.fun > self.config['thresholds'][0]:
                #if check fails return false
                return False
            # check if all loss functions are under the corresponding threshold
            for ii in range(1, len(lossfunctions)):
                if lossfunctions[ii](result.x) > self.config['thresholds'][ii]:
                    # if check fails return false
                    return False
        # when no check fails return True  = success
        return True

    def weights_to_valid_input(self, weights: list) -> list:
        """
        need to change weights from scip optimizer to proper input 
        if one optimize with complexe values

        Parameters
        ----------
        weights : list
            list of weights

        Raises
        ------
        ValueError
            if imaginary is not defined correctly in config file

        Returns
        -------
        list
            ordered list according to imaginary

        """
        if self.imaginary ==  'cartesian':

            if len(weights) % 2 == 0:
                ll2 = int(len(weights)/2)
            else:
                raise ValueError(
                    'odd number of weights for complex optimization')

            return [complex(real, imag) for real, imag in
                    zip(weights[:ll2], weights[ll2:])]

        elif self.imaginary == 'polar':

            if len(weights) % 2 == 0:
                ll2 = int(len(weights)/2)
            else:
                raise ValueError(
                    'odd number of weights for complex optimization')
            return [(radius, phase) for radius, phase in
                    zip(weights[:ll2], weights[ll2:])]
        elif self.imaginary is False:
            return weights
        else:
            raise ValueError('imaginary: only: polar,cartesian,false')

    def get_loss_functions(self, current_graph: Graph):
        """
        get a list of all loss functions mentioned in config

        Parameters
        ----------
        current_graph

        Returns
        -------
        callable_loss
            list of callable functions

        """
        # get loss function acc. to config
        lossfunctions = loss_dic[self.config['loss_func']]

        # entanglement loss function
        if self.config['loss_func'] == 'ent':  # we optimize for entanglement
            loss_specs = {'sys_dict': self.ent_dic,
                          'imaginary': self.imaginary,
                          'var_factor': self.config['var_factor']}

        # CR and FID
        elif self.config['loss_func'] in ['cr', 'fid']:
            loss_specs = {'target_state': self.target,
                          'cnfg': self.config}

        # fock basis
        elif self.config['loss_func'] in ['fockcr','fockfid']:
            #loss_specs = {'target_state': self.target,
            #              'cnfg': self.config}
            loss_specs = {'target_state': self.target,
                        'num_anc':self.config['num_anc'],
                        'amplitudes':self.config['amplitudes'],
                        'imaginary': self.imaginary}

        # custom loss functions
        elif self.config['loss_func'] == 'lff':
            loss_specs = {'cnfg': self.config}
        callable_loss = [func(current_graph, **loss_specs)
                         for func in lossfunctions]
        return callable_loss

    def update_losses(self, result, losses):
        """
        updates the losses for next steps

        Parameters
        ----------
        result : scipy.minimizer object
            minimize object from optimizaiton step
        losses : list
            list of loss functions

        Returns
        -------
        list of losses for corrosponding weights stored in result.x

        """
        loss_values = [result.fun]
        for ii in range(1, len(losses)):
            loss_values.append(losses[ii](result.x))
        return loss_values

    def pre_optimize_start_graph(self, graph) -> Graph:
        """
        first optimization of complete starting graph

        Parameters
        ----------
        graph

        Returns
        -------
        preopt_graph

        """
        # losses is a list of callable lossfunctions, e.g. [countrate(x), fidelity(x)], where x is a vector of edge weights
        # that can be given to scipy.optimize
        losses = self.get_loss_functions(graph)
        valid = False
        counter = 0
        # repeat optimization of complete graph until a good solution is found (which satifies self.check())
        while not valid:
            # prepare optimizer
            initial_values, bounds = self.prepOptimizer(len(graph))
            # optimization with scipy
            best_result = optimize.minimize(losses[0], x0=initial_values,
                                            bounds=bounds,
                                            method=self.config['optimizer'],
                                            options={'ftol': self.config['ftol']})
            self.loss_val = self.update_losses(best_result, losses)
            #check if solution is valid
            valid = self.check(best_result, losses)
            counter += 1
            #print a warning if preoptimization is stuck in a loop
            if counter % 10 == 0:
                print('10 invalid preoptimization, consider changing parameters.')
                log.info('10 invalid preoptimization, consider changing parameters.')

        # if num_pre is set to larger than 1 in config, do num_pre preoptimization and choose the best one.
        # for optimizations with concrete target state, num_pre = 1 is enough
        for __ in range(self.config['num_pre'] - 1):
            initial_values, bounds = self.prepOptimizer(len(graph))
            result = optimize.minimize(losses[0], x0=initial_values,
                                       bounds=bounds,
                                       method=self.config['optimizer'],
                                       options={'ftol': self.config['ftol']})

            if result.fun < best_result.fun:
                best_result = result
        self.loss_val = self.update_losses(best_result, losses)
        print(f'best result from pre-opt: {abs(best_result.fun)}')
        log.info(f'best result from pre-opt: {abs(best_result.fun)}')

        preopt_graph = Graph(graph.edges,
                             weights=self.weights_to_valid_input(
                                 best_result.x),
                             imaginary=self.imaginary)

        try:
            bulk_thr = self.config['bulk_thr']
        except:
            bulk_thr = 0
        if bulk_thr > 0:
            # cut all edges smaller than bulk_thr and optimize again
            # this can save a lot of time
            cont = True
            num_deleted = 0
            while cont:
                # delete smallest edges one by one
                idx_of_edge = preopt_graph.minimum()
                amplitude = preopt_graph[idx_of_edge]
                if self.imaginary == 'polar':
                    amplitude = amplitude[0]
                if abs(amplitude) < bulk_thr:
                    preopt_graph.remove(idx_of_edge)
                    num_deleted += 1
                else:
                    cont = False
            print(f'{num_deleted} edges deleted')
            log.info(f'{num_deleted} edges deleted')
            valid = False
            while not valid:
                # it is necessary that the truncated graph passes the checks
                initial_values, bounds = self.prepOptimizer(len(preopt_graph))
                losses = self.get_loss_functions(preopt_graph)
                trunc_result = optimize.minimize(losses[0], x0=initial_values,
                                                 bounds=bounds,
                                                 method=self.config['optimizer'],
                                                 options={'ftol': self.config['ftol']})
                self.loss_val = self.update_losses(trunc_result, losses)
                print(f'result after truncation: {abs(trunc_result.fun)}')
                log.info(f'result after truncation: {abs(trunc_result.fun)}')
                valid = self.check(trunc_result, losses)
            preopt_graph = Graph(preopt_graph.edges,
                                 weights=self.weights_to_valid_input(
                                     trunc_result.x),
                                 imaginary=self.imaginary)

        return preopt_graph

    def prepOptimizer(self, numweights, x=[]):
        """
        returns initial values and bounds for use in optimization.

        Parameters
        ----------
        numweights
        x

        Returns
        -------
        initial_values
        bounds
        """

        if not self.imaginary:
            bounds = numweights * [(-1, 1)]
            if len(x) == 0:
                initial_values = 2 * np.random.random(numweights) - 1
            else:
                initial_values = x
        else:
            bounds = numweights * [(-1, 1)] + numweights * [(-np.pi, np.pi)]
            if len(x) == 0:
                rands_r = 2 * np.random.random(numweights) - 1
                rands_th = 2 * np.pi * np.random.random(numweights) - np.pi
                initial_values = np.concatenate([rands_r, rands_th])
            else:
                initial_values = x

        return initial_values, bounds

    def continuationCondition(self, num_edge) -> bool:
        """
        conditions that stop optimization

        Parameters
        ----------
        num_edge

        Returns
        -------
        cont: bool
            if True, topological optimization is continued
        """
        if self.config['loss_func'] == 'ent':
            cont = len(self.graph) > self.config['min_edge'] and num_edge < len(
                self.graph)
        else:
            # if num_edge is higher than total number of edges in the graph
            # or higher than edges_tried, return False
            cont = num_edge < min(len(self.graph), self.config['edges_tried'])
        return cont

    def optimize_one_edge(self, num_edge: int,
                          num_tries_one_edge: int) -> (Graph, bool):
        """
        delete the num_edge-th smallest edge and optimize num_tries_one_edge times
        and check if corresponding loss function fulfills checks

        Parameters
        ----------
        num_edge
        num_tries_one_edge

        Returns
        -------
        new_graph
           if edge is successfully deleted, reduced graph is returned. else graph is not modified.
        success
        """
        # copy current Graph and delete num_edge´s smallest weight
        # set up reduced graph
        reduced_graph = self.graph.copy()
        #find index of num_edge smallest edge
        idx_of_edge = reduced_graph.minimum(num_edge)
        #store amplitude in case edge fails and needs to be put back in
        amplitude = reduced_graph[idx_of_edge]
        #remove smallest edge
        reduced_graph.remove(idx_of_edge, update=False)
        #update state catalog of reduced graph
        reduced_graph.getStateCatalog()
        #try a given number of times to delete this edge
        for ii in range(num_tries_one_edge):
            # if edge is tried for the first time, update loss function and other optimization parameters
            # also the weights of the old graph are used as initial values (this is much faster)
            if ii == 0:
                try:
                    # redefine loss functions for reduced graph
                    losses = self.get_loss_functions(reduced_graph)
                    # use initial values x0 from previous Graph
                    x0 = reduced_graph.weights
                    initial_values, bounds = self.prepOptimizer(len(reduced_graph),
                                                                x=x0)
                    # optimize with scipy
                    result = optimize.minimize(losses[0], x0=initial_values,
                                               bounds=bounds,
                                               method=self.config['optimizer'],
                                               options={'ftol': self.config['ftol']})
                except KeyError:
                    # if the target kets can not be produced with the given graph we can give up on this edge
                    # it wont work
                    reduced_graph[idx_of_edge] = amplitude
                    print('edge necessary for producing all kets')
                    log.info('edge necessary for producing all kets')
                    return reduced_graph, False  # no success keep current Graph
            # if edge has been tried before, reuse loss function etc. and use random initial values
            else:
                # random initial values
                initial_values, bounds = self.prepOptimizer(len(reduced_graph))
                # optimize with scipy
                result = optimize.minimize(losses[0], x0=initial_values,
                                           bounds=bounds,
                                           method=self.config['optimizer'],
                                           options={'ftol': self.config['ftol']})
            # check if solution is valid
            valid = self.check(result, losses)

            if valid:  # if criterion is reached then save reduced graph as graph, else leave graph as is
                # compute values for all losses
                self.loss_val = self.update_losses(result, losses)
                # if clean solution is encountered before optimization finishes, save that too
                # all weights are +-1 and solution is pretty good, it is interesting enough to be saved
                # to a file even if topological optimization can be continued
                if all(np.array(abs(self.graph)) > 0.95):
                    self.saver.save_graph(self)
                if self.save_hist:
                    self.history.append(self.loss_val)
                # return updated result graph
                return Graph(reduced_graph.edges,
                             weights=self.weights_to_valid_input(result.x),
                             imaginary=self.imaginary), True
        # all tries failed keep current Graph
        reduced_graph[idx_of_edge] = amplitude
        return reduced_graph, False

    def topologicalOptimization(self, save_hist=True) -> Graph:
        """
        does the topological main loop. deletes edges until continuation condition fails.
        Returns
        -------
        solution_graph
            result of optimization
        """
        #start with smallest edge
        num_edge = 0
        graph_history = []
        # if num_edge becomes too large, optimization is stopped
        while self.continuationCondition(num_edge):
            # try if num_edge smallest edge can be removed
            # if successful, return optimized graph after deletion
            # if not successful, return old graph
            self.graph, success = self.optimize_one_edge(
                num_edge, self.config['tries_per_edge'])
            #iterate num_edge to try next smallest edge
            num_edge += 1
            log.info(f'deleting edge {num_edge}')
            print(f'deleting edge {num_edge}')
            if success:
                print(
                    f"deleted: {len(self.graph)}  edges left with loss {self.loss_val[0]:.3f}")
                log.info(f"deleted: {len(self.graph)}  edges left with loss {self.loss_val[0]:.3f}")
                #reset to try smallest edge again for next iteration
                num_edge = 0
                graph_history.append(self.graph)

        return self.graph



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 14 17:15:59 2022

@author: janpe
"""
import os.path

from .fancy_classes import Graph
import numpy as np
from pathlib import Path
import json


def read_json(abspath: Path) -> dict:
    """
    return the dictonary for a given abspath that is a json file

    """
    with abspath.open('r', encoding="UTF-8") as openfile:
        # Reading from json file
        dictonary = json.load(openfile)
        openfile.close()

    return dictonary


def write_json(abspath: Path, dictionary: dict,
               overwrite_existing_file = False) -> None:
    """
    getting an Path object and a dictonary then write the dictionary
    in a json file for the given path

    """

    # Serializing json
    class NumpyEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, np.ndarray):
                return obj.tolist()
            if isinstance(obj, np.int32):
                return int(obj)
            if isinstance(obj, np.int64):
                return int(obj)
            return json.JSONEncoder.default(self, obj)

    if not isinstance(abspath, Path):
        abspath = Path(abspath)

    # convert json object
    json_object = json.dumps(dictionary, cls=NumpyEncoder,
                             ensure_ascii=False, indent=4)

    ### write json to new folder ###
    # make json ending
    abspath_save = str(abspath)
    if not abspath_save.endswith('.json'):
        abspath_save += '.json'
    # check if same state exists
    if overwrite_existing_file is False:
        counter = 0
        while Path(abspath_save).exists():
            abspath_save = str(abspath) + f'({counter})' + '.json'
            counter += 1
        
    abspath = Path(abspath_save)
    with abspath.open("w+", encoding="UTF-8") as file:
        file.write(json_object)
        file.close()


class saver:

    def __init__(self, config=None, name_config_file = '', dim = []):

        self.name_config_file = Path(name_config_file).name
        self.config = config
        self.config['dimensions'] = dim
        self.best_state = None
        self.save_path = self.get_and_create_save_directory()
        self.best_opt = None

    def get_folder_name(self) -> str:
        """
        return the folder name: dimension of the system or confi

        """
        folder = self.name_config_file
        if self.name_config_file.endswith('.json'):
            folder = self.name_config_file[:-5]
        return Path(folder) / self.config.get('foldername', 'try')

    def get_and_create_save_directory(self):
        """
        look if folder ~/output/(2-2-2-2) exists otherwise creates it
        when exists it looks if config files are the same then our
        safe directory is ~/output/(2-2-2-2) otherwise: we chose
        ~/data/(2-2-2-2)(x) for a x where the summary files are matching
        or a new directory
        """
        folder_name = self.get_folder_name()

        i = 0
        while True:  # iterate as long as one could find a proper safe folder
            pt = Path.cwd() / 'output' / folder_name  # move data directory
            pt.mkdir(parents=True, exist_ok=True)
            summary_path = pt / 'summary.json'
            if summary_path.exists():
                if self.check_if_summary_is_same(summary_path):
                    self.foldername = Path(folder_name)
                    return pt
                else:  # adapt folder name to avoid same folder with different summaries
                    folder_name = str(self.get_folder_name()) + f'_{i}'
                    i += 1
            else:
                self.foldername = Path(folder_name)
                write_json(pt / 'summary.json', self.config)
                return pt

    def check_if_summary_is_same(self, path_exst_summary: Path) -> bool:
        """
        just checks if the existing summary and the current have the
        same parameters, the compare function is needed for
        distinguishing iterable and non iterables
        """

        def compare(obj_a, obj_b):
            try:
                __ = iter(obj_a)
                return all([a == b for a, b in zip(obj_a, obj_b)])
            except TypeError:
                return obj_a == obj_b

        existing_summary = read_json(path_exst_summary)
        try:
            same = all([compare(self.config[key], existing_summary[key])
                        for key in self.config.keys() if key != 'samples'] )
        except KeyError:
            same = False
        return same

    def convert_graph_keys_in_str(self, graph: dict) -> dict:
        """
        here we can convert our Graph dict in a dict that has strings as keys
        we need this to save it in json file

        """
        # convert keys in str
        ret_dict = {}
        for key in graph.keys():
            if type(key) is not str:
                try:
                    ret_dict[str(key)] = graph[key]
                except:
                    try:
                        ret_dict[repr(key)] = graph[key]
                    except:
                        pass

        return ret_dict

    def check_best_opt(self, topo):
        # check if opt is better than previous best_opt
        # this may not only depend on the loss function. for example, a solution with fidelity 0.994 and 16 edges
        # is 'better' in some cases than fidelity 0.999 and 23 edges
        # TODO: implement the choice of which criteria is used
        return self.best_opt is None or topo.loss_val < self.best_opt.loss_val
    
    def get_dictonary_storing_in_json(self,topo_obj):
        """
        converts the topological optimizer object to a dict

        Parameters
        ----------
        topo_obj : Object of topological_opti

        Returns
        -------
        safe_dic : dict

        """
        
        
        safe_dic = {'graph': self.convert_graph_keys_in_str(topo_obj.graph.graph),
                    'loss': topo_obj.loss_val,
                    'seed': self.config['seed'],
                    'history': topo_obj.history}

        try:
            safe_dic['graph_hist'] = [self.convert_graph_keys_in_str(xx.graph)
                                      for xx in topo_obj.graph_hist]
            safe_dic['loss_hist'] = topo_obj.loss_hist
        except AttributeError:
            pass
        return safe_dic
    
    def save_graph(self, topo: object) -> None:
        """
        we use an object from the class topological_opti to save all
        infos: - the optimized graph
               - corresponding loss
               - if confi.safe_hist is True we also safe the loss during
                 each deletion and the corresponding graph

        """

        abs_path = self.save_path / self.get_file_name(topo.graph, topo.loss_val)
        # update best graph
        if self.check_best_opt(topo):
            self.best_opt = topo
            safe_dic_best = self.get_dictonary_storing_in_json(topo)
            print('new best state saved to best.json')
            write_json(self.save_path / 'best' , safe_dic_best, 
                       overwrite_existing_file = True)
            # TODO: rewrite best_graph to file on this line (then we get best graph even if run is cancelled before it is finished)
            
        safe_dic = self.get_dictonary_storing_in_json(topo)

        write_json(abs_path, safe_dic)

    def get_file_name(self, graph: Graph, loss: float) -> str:
        """
        takes as input object Graph and generate a file name out of it

        """
        if all(np.array(abs(graph)) > 0.95):
            clean = 'clean'
        else:
            clean = 'rough'
        file_name = clean + '-' + str(len(graph.graph)) + '-'
        graph.getStateCatalog()
        file_name += str(len(graph.perfect_matchings)) + '-'
        for lo in loss:
            file_name += f'{lo:.4f}_'
        return file_name[:-1] # delet last _



import numpy as np
import pytheus.theseus as th
from copy import deepcopy

# +
DEFAULT_NORM = 'Not stored yet. You have to run getNorm().'
DEFAULT_STATE = 'Not stored yet. You have to run getState().'
WRONG_IMAGINARY = 'The property `imaginary` is NOT defined correctly.'


def invalidInput(input_name):
    return f'Introduce a valid input `{input_name}`.'


def defaultValues(length, imaginary):
    if imaginary in [False, 'cartesian']:
        return [True] * length
    elif imaginary == 'polar':
        return [(True, False)] * length
    else:
        raise ValueError(invalidInput('imaginary'))


# -

# # The Graph class

class Graph():  # should this be an overpowered dictionary? NOPE
    def __init__(self,
                 edges,  # list/tuple of edges, dictionary with weights, 'full' or 'empty'
                 dimensions=[],
                 weights=[],  # list of values or tuples encoding imaginary values
                 imaginary=False,  # 'cartesian' or 'polar'
                 norm=False,  # For the sake of perfomance, compute
                 state=False,  # norm and state only when needed.
                 state_cat=True
                 ):
        self.dimensions = dimensions
        self.imaginary = imaginary
        self.full = True if edges == 'full' else False
        # The next line may redefine previous properties
        self.graph = self.graphStarter(edges, weights)  # MAIN PROPERTY
        # This may not be elegant, but it works
        self.state_catalog = None
        if self.state_catalog == None and state_cat: self.getStateCatalog()
        self.norm = DEFAULT_NORM
        if norm: self.getNorm()
        self.state = DEFAULT_STATE
        if state: self.getState()

    # Long, cumbersome, and very necessary function.
    def graphStarter(self, edges, weights):
        '''
        To facilitate the use of the library, this function transforms different kinds of graph
        inputs into a suitable Graph instance, redefining the properties if needed.
        
        Once the edges and weights are correct, it returns an property graph.
        '''
        if edges == 'full':  # If True, dimensions must be defined.
            if len(self.dimensions) == 0:
                raise ValueError('Introduce the dimensions to get a fully connected graph')
            else:
                edges = th.buildAllEdges(self.dimensions)
                if len(weights) == 0:
                    weights = defaultValues(len(edges), self.imaginary)
                else:
                    raise ValueError('The input `full` does not allow to specify the weights.')
        # If the introduced edges are a list/tuple, they may include the weights           
        elif type(edges) in [list, tuple]:
            edge_shape = np.shape(edges)
            if edge_shape[1] == 4:
                pass  # seems legit
            elif edge_shape[1] == 5:  # The weights will be the last term of each item
                weights = [ed[-1] for ed in edges]
                edges = [ed[:4] for ed in edges]
            elif edge_shape[1] == 6:  # The weights will be the last 2 terms of each item
                weights = [tuple(ed[-2], ed[-1]) for ed in edges]
                edges = [ed[:4] for ed in edges]
            else:
                raise ValueError(invalidInput('edges'))
        elif type(edges) == dict:
            weights = [edges[key] for key in sorted(edges.keys())]
            edges = [kk if isinstance(kk, tuple) else eval(kk) for kk in sorted(edges.keys())]

            # Verification of appropiate edges
            if all(isinstance(edge, tuple) for edge in edges):
                if all(len(edge) == 4 for edge in edges):
                    pass  # seems legit
                else:
                    raise ValueError(invalidInput('edges'))
            else:
                raise ValueError(invalidInput('edges'))
        else:
            raise ValueError(invalidInput('edges'))

        if len(weights) == 0:  # The default option True behaves (mostly) as 1
            weights = defaultValues(len(edges), self.imaginary)
        else:
            weight_shape = np.shape(weights)
            if weight_shape[0] != len(edges):
                raise ValueError('The number of edges and weights do not match.')
            elif len(weight_shape) == 1:
                if any(isinstance(val, complex) for val in weights):
                    self.imaginary = 'cartesian'
            elif len(weight_shape) == 2:
                if weight_shape[1] == 2:
                    if self.imaginary == 'cartesian':
                        weights = [val[0] + 1j * val[1] for val in weights]
                    elif self.imaginary == 'polar':
                        weights = [tuple(val) for val in weights]
                    else:
                        raise ValueError(invalidInput('imaginary'))
                else:
                    raise ValueError('Introduce valid weights.')
            else:
                raise ValueError('Introduce valid weights.')

        if len(self.dimensions) == 0:
            self.dimensions = th.graphDimensions(edges)  # a nice function should be defined here
        # Once edges and weights are properly defined, we return a graph
        return {ed: val for ed, val in zip(edges, weights)}

    def __repr__(self):
        '''
        What you see when using print() or calling the instance.
        '''
        return '{' + ',\n '.join(f'{kk}: {vv}' for kk, vv in self.graph.items()) + '}'

    def __len__(self):
        return len(self.graph)

    def __eq__(self, other):
        return self.graph == other.graph

    def __getitem__(self, edge):
        return self.graph[tuple(edge)]

    def __setitem__(self, edge, weight):
        if isinstance(edge, (tuple, list)) and len(edge) == 4:
            if isinstance(weight, (int, float, complex)):
                if self.imaginary == 'polar':
                    self.graph[tuple(edge)] = (abs(weight), np.angle(weight))
                    print('Weight stored in polar notation.')
                else:
                    self.graph[tuple(edge)] = weight
                    if type(weight) == complex:
                        self.imaginary = 'cartesian'
            elif isinstance(weight, (tuple, list)) and len(weight) == 2:
                if self.imaginary == 'polar':
                    self.graph[edge] = tuple(weight)
                else:
                    self.graph[edge] = weight[0] * np.exp(1j * weight[1])
                    self.imaginary = 'cartesian'
                    print('Weight stored in cartesian notation.')
            else:
                raise ValueError(invalidInput('weight'))
        else:
            raise ValueError(invalidInput('edge'))

    # The method __round__ for the class State is almost the same, could it be reformat?
    def __round__(self, ndigits=0):
        if self.imaginary == False:
            for edge in self.edges:
                self[edge] = round(self[edge], ndigits)
        elif self.imaginary == 'cartesian':
            for edge in self.edges:
                self[edge] = round(self[edge].real, ndigits) + 1j * round(self[edge].imag, ndigits)
        elif self.imaginary == 'polar':
            for edge in self.edges:
                self[edge] = (round(self[edge][0], ndigits), self[edge][1])
        else:
            raise ValueError(WRONG_IMAGINARY)
        return self

    # this should be __del__ but then you would have to do: del self.graph[] etc
    def remove(self, edge, update=True):
        del self.graph[edge]
        if update:
            remove_ket_list = []
            for ket, pm_list in self.state_catalog.items():
                if ((edge[0], edge[2]) and (edge[1], edge[3])) in ket:
                    self.state_catalog[ket] = [pm for pm in pm_list if edge not in pm]
                    if len(self.state_catalog[ket]) == 0:
                        remove_ket_list.append(ket)
            for ket in remove_ket_list:
                del self.state_catalog[ket]
        # if update:
        #    if self.norm != DEFAULT_NORM: self.getNorm()
        #    if self.state != DEFAULT_STATE: self.getState()

    def purge(self, threshold=1e-4, update=True):
        '''
        It removes all edges whose weights, in absolute value, are below `threshold`.
        '''
        remove_edges = []
        if self.imaginary == 'polar':
            for edge, weight in self.graph.items():
                if abs(weight[0]) < threshold:
                    remove_edges.append(edge)
        else:
            for edge, weight in self.graph.items():
                if abs(weight) < threshold:
                    remove_edges.append(edge)
        for edge in remove_edges:
            del self.graph[edge]
        if update:
            remove_ket_list = []
            for ket, pm_list in self.state_catalog.items():
                for edge in remove_edges:
                    if ((edge[0], edge[2]) and (edge[1], edge[3])) in ket:
                        self.state_catalog[ket] = [pm for pm in pm_list if edge not in pm]
            for ket, pm_list in self.state_catalog.items():
                if len(pm_list) == 0:
                    remove_ket_list.append(ket)
            for ket in remove_ket_list:
                del self.state_catalog[ket]

    # DEFINE GET, SET AND DEL ITEM

    @property
    def edges(self):
        return list(self.graph.keys())

    @property
    def weights(self):
        return list(self.graph.values())

    @property
    def num_nodes(self):
        return len(self.dimensions)

    @property
    def perfect_matchings(self):
        return sum(self.state_catalog.values(), [])

    @property
    def is_weighted(self):
        if all(isinstance(val, tuple) for val in self.weights):
            return not all(isinstance(val, bool) for val in sum(self.weights, ()))
        elif all(isinstance(val, bool) for val in self.weights):
            return False
        elif all(isinstance(val, (int, float, complex)) for val in self.weights):
            return True
        else:
            raise ValueError('The weights are NOT defined correctly.')

    # imaginary could be redefined as one of these properties

    def copy(self):
        return deepcopy(self)

    def addEdge(self, edge, weight=None, update=True):
        if len(edge) == 4 and all(isinstance(val, int) for val in edge):
            edge = tuple(edge)
        else:
            raise ValueError(invalidInput('edge'))
        if edge in self.edges:
            print('The edge is going to be redefined.')
            update = False  # there's no need to update the state_catalog
        else:  # the edge is included
            self.graph[edge] = True
        if weight == None:
            self[edge] = defaultValues(1, self.imaginary)[0]
        else:
            self[edge] = weight
        if update:
            subgraph = th.removeNodes(edge[:2], self.edges)
            new_states = th.stateCatalog(th.findPerfectMatchings(subgraph + [edge]))
            for ket, perfect_matchings in new_states.items():
                try:
                    self.state_catalog[ket] += perfect_matchings
                except KeyError:
                    self.state_catalog[ket] = perfect_matchings
                self.state_catalog[ket] = sorted(self.state_catalog[ket])
        #     if self.norm != DEFAULT_NORM: self.getNorm()
        # if self.state != DEFAULT_STATE: self.getState()

    # This could be also a property, but then we cannot introduce arguments
    def node_degrees(self, ascending=False):  #
        '''
        Degree of each node of the graph.
        '''
        return th.nodeDegrees(self.edges, rising=ascending)

    def getStateCatalog(self):
        if self.full:
            self.state_catalog = th.allPerfectMatchings(self.dimensions)
        else:
            self.state_catalog = th.stateCatalog(th.findPerfectMatchings(self.edges))

    # the employed norm function is simplified (only pm) and uses the new imaginary notation
    def getNorm(self):
        self.norm = th.writeNorm(self.state_catalog, imaginary=self.imaginary)

    def getState(self):
        kets = list(self.state_catalog.keys())
        if self.is_weighted:
            amplitudes = []
            conversion = self.imaginary == 'polar'
            if conversion:
                self.toCartesian()
            for kt in kets:
                term = 0
                for pm in self.state_catalog[kt]:
                    term += np.prod([self.graph[edge] for edge in pm])
                amplitudes.append(term)
            if conversion:
                self.toPolar()
        else:
            amplitudes = []
        self.state = State(kets, amplitudes, self.imaginary)

        # This could also be defined as __abs__, but what do you give back? The dictionary?

    def __abs__(self):
        return_weights = len(self.graph) * [0]
        if self.is_weighted:
            if self.imaginary in [False, 'cartesian']:
                for idx, vv in enumerate(self.graph.values()):
                    return_weights[idx] = abs(vv)
            elif self.imaginary == 'polar':
                for idx, vv in enumerate(self.graph.values()):
                    return_weights[idx] = abs(vv[0])
            else:
                raise ValueError(WRONG_IMAGINARY)
            return return_weights
        else:
            ValueError('emtpy weights')

    def absolute(self):
        if self.is_weighted:
            if self.imaginary in [False, 'cartesian']:
                for kk, vv in self.graph.items():
                    self.graph[kk] = abs(vv)
            elif self.imaginary == 'polar':
                for kk, vv in self.graph.items():
                    self.graph[kk] = abs(vv[0])
            else:
                raise ValueError(WRONG_IMAGINARY)
        else:
            for kk, vv in self.graph.items():
                self.graph[kk] = True
        self.imaginary = False

    def toCartesian(self):
        if self.imaginary == 'cartesian':
            print('The weights are already in cartesian coordinates.')
            return None
        if self.is_weighted:
            if self.imaginary == False:
                for kk, vv in self.graph.items():
                    self.graph[kk] = (vv, 0)
            elif self.imaginary == 'polar':
                for kk, vv in self.graph.items():
                    self.graph[kk] = vv[0] * np.exp(1j * vv[1])
            else:
                raise ValueError('The propery `imaginary` is NOT defined correctly.')
        else:
            for kk, vv in self.graph.items():
                self.graph[kk] = True
        self.imaginary = 'cartesian'

    def toPolar(self):
        if self.imaginary == 'polar':
            print('The weights are already in polar coordinates.')
            return None
        if self.is_weighted:
            if self.imaginary == False:
                for kk, vv in self.graph.items():
                    self.graph[kk] = (vv, 0)
            elif self.imaginary == 'cartesian':
                for kk, vv in self.graph.items():
                    self.graph[kk] = (abs(vv), np.angle(vv))
            else:
                raise ValueError('The propery `imaginary` is NOT defined correctly.')
        else:
            for kk, vv in self.graph.items():
                self.graph[kk] = (True, False)
        self.imaginary = 'polar'

    def rescale(self, constant):
        conversion = self.imaginary == 'polar'
        if conversion:
            self.toCartesian()
        for edge in self.graph.keys():
            self.graph[edge] *= constant
        if conversion:
            self.toPolar()

    def minimum(self, *args):
        '''
        returns the key of the edge with the smallest weight, e.g:
            graph.minimum() smallest edge
            graph.minimum(-1) biggest edge
            graph.minimum(0,5) first 5 smallest edges 
        '''
        if len(args) == 0:
            n_th_smallest = 0  # takes smallest
        elif len(args) == 1:
            n_th_smallest = args[0]  # takes nth given
        else:
            n_th_smallest = slice(*args)

        if self.imaginary in [False, 'cartesian']:
            idx = np.argsort(abs(np.array(self.weights)))
        elif self.imaginary == 'polar':
            idx = np.argsort(abs(np.array([rr[0] for rr in self.weights])))
        # now check if input is valid
        try:
            delind = idx[n_th_smallest]
        except IndexError:
            lenght_graph = len(self.graph)
            max_given_n = max(args)
            raise ValueError(
                f'Given n_th is to large (n starts 0): {max_given_n+1=} > {lenght_graph=}')

        if type(delind) is np.int64:  # makes sure that we can iterate by return
            return self.edges[delind]

        return [self.edges[ii] for ii in delind]

    def clamp(self, maximum=1, minimum=None):  # , rescale=False):
        if maximum >= 0:
            if minimum == None:
                minimum = - maximum
            for edge, weight in self.graph.items():
                self.graph[edge] = max(minimum, min(weight, maximum))
        else:
            raise ValueError('Introduce a positive maximum.')


# # The State class

class State():

    def __init__(self,
                 kets,
                 amplitudes=[],  # list of values or tuples encoding imaginary values
                 imaginary=False,  # 'cartesian' or 'polar'
                 normalize=False,
                 ):

        self.imaginary = imaginary
        self.state = self.stateStarter(kets, amplitudes)  # This may redefine previous properties
        if normalize: self.normalize()

    def stateStarter(self, kets, amplitudes):
        '''
        Function to initiate a State instance with different inputs.
        This version is not so flexible with the input format as the analogous from Graph.
        '''
        if type(kets) == dict:
            amplitudes = [kets[key] for key in sorted(kets.keys())]
            kets = sorted(kets.keys())
        # Here, introducing the amplitudes after the kets in a list/tuple doesn't work
        elif type(kets) in [list, tuple]:
            pass  # seems legit
        else:
            raise ValueError(invalidInput('kets'))

        # Verification of appropiate kets            
        kets_shape = np.shape(kets)
        if all(isinstance(kt, str) for kt in kets):
            kets = [tuple((ii, int(dim)) for ii, dim in enumerate(kt)) for kt in kets]
        elif kets_shape[2] == 2:
            pass  # The third component must have 2 dimensions: (node, dim)
        else:
            raise ValueError('Introduce valid input `kets`.')

        # Verification and setting of appropiate amplitudes
        if len(amplitudes) == 0:  # The default option True behaves (mostly) as 1
            amplitudes = defaultValues(len(kets), self.imaginary)
        else:
            amplitude_shape = np.shape(amplitudes)
            if amplitude_shape[0] != len(kets):
                raise ValueError('The number of kets and amplitudes do not match.')
            elif len(amplitude_shape) == 1:
                if any(isinstance(val, complex) for val in amplitudes):
                    self.imaginary = 'cartesian'
            elif len(amplitude_shape) == 2:
                if amplitude_shape[1] == 2:
                    if self.imaginary == 'cartesian':
                        amplitudes = [val[0] + 1j * val[1] for val in amplitudes]
                    elif self.imaginary == 'polar':
                        amplitudes = [tuple([val[0], val[1]]) for val in amplitudes]
                    else:
                        raise ValueError(invalidInput('imaginary'))
                else:
                    raise ValueError('Introduce valid amplitudes.')
            else:
                raise ValueError('Introduce valid amplitudes.')

        # Once kets and amplitudes are properly defined, we return a state
        return {kt: val for kt, val in zip(kets, amplitudes)}

    def __repr__(self):
        '''
        What you see when using print() or calling the instance.
        '''
        return '{' + ',\n '.join(f'{kk}: {vv}' for kk, vv in self.state.items()) + '}'

    def __len__(self):
        return len(self.state)

    # def __eq__(self, other):
    # Options for establishing equivalence:
    # - Direct comparison of dictionaries: count kets as existing even if amplitude 0
    # - Using the inner product, but a@b=-.9999, is it equal? threshold needed 

    def __getitem__(self, ket):
        if type(ket) == str:
            ket = tuple((ii, int(dim)) for ii, dim in enumerate(ket))
        return self.state[ket]

    def __setitem__(self, ket, amplitude):
        if isinstance(ket, str):
            ket = tuple((ii, int(dim)) for ii, dim in enumerate(ket))
        elif isinstance(ket, (tuple, list)):
            pass  # seems legit
        else:
            raise ValueError(invalidInput('ket'))

        if isinstance(amplitude, (int, float, complex)):
            if self.imaginary == 'polar':
                self.state[tuple(ket)] = (abs(amplitude), np.angle(amplitude))
                print('Amplitude stored in polar notation.')
            else:
                self.state[tuple(ket)] = amplitude
                if type(amplitude) == complex:
                    self.imaginary = 'cartesian'
        elif isinstance(amplitude, (tuple, list)) and len(amplitude) == 2:
            if self.imaginary == 'polar':
                self.state[ket] = tuple(amplitude)
            else:
                self.state[ket] = amplitude[0] * np.exp(1j * amplitude[1])
                self.imaginary = 'cartesian'
                print('Amplitude stored in cartesian notation.')
        else:
            raise ValueError(invalidInput('amplitude'))

    def __matmul__(self, other):
        '''
        Braket operation, the inner product between state. 
        It can be used with @.
        '''
        conversion_a = self.imaginary == 'polar'
        if conversion_a:
            self.toCartesian()
        conversion_b = other.imaginary == 'polar'
        if conversion_b:
            other.toCartesian()
        amplitudes_a = []
        amplitudes_b = []
        for ket in set(self.kets + other.kets):
            try:
                amplitudes_a.append(self.state[ket])
            except KeyError:
                amplitudes_a.append(0)
            try:
                amplitudes_b.append(other.state[ket])
            except KeyError:
                amplitudes_b.append(0)
        if conversion_a:
            self.toPolar()
        if conversion_b:
            other.toPolar()
        # turning into array the second term is redundant, but somehow faster
        return np.conjugate(amplitudes_a) @ np.array(amplitudes_b)

    # The method __round__ for the class State is almost the same, could it be reformat?
    def __round__(self, ndigits=0):
        if self.imaginary == False:
            for ket in self.kets:
                self[ket] = round(self[ket], ndigits)
        elif self.imaginary == 'cartesian':
            for ket in self.kets:
                self[ket] = round(self[ket].real, ndigits) + 1j * round(self[ket].imag, ndigits)
        elif self.imaginary == 'polar':
            for ket in self.kets:
                self[ket] = (round(self[ket][0], ndigits), self[ket][1])
        else:
            raise ValueError(WRONG_IMAGINARY)
        return self

    def purge(self, threshold=1e-4):
        '''
        It removes all kets whose amplitudes, in absolute value, are below `threshold`.
        '''
        remove_kets = []
        if self.imaginary == 'polar':
            for ket, amplitude in self.state.items():
                if abs(amplitude[0]) < threshold:
                    remove_kets.append(ket)
        else:
            for ket, amplitude in self.state.items():
                if abs(amplitude) < threshold:
                    remove_kets.append(ket)
        for ket in remove_kets:
            del self.state[ket]

    @property
    def kets(self):
        return list(self.state.keys())

    @property
    def amplitudes(self):
        return list(self.state.values())

    @property
    def norm(self):
        return (self @ self) ** 0.5

    @property
    def is_weighted(self):
        if all(isinstance(val, tuple) for val in self.amplitudes):
            return not all(isinstance(val, bool) for val in sum(self.amplitudes, ()))
        elif all(isinstance(val, bool) for val in self.amplitudes):
            return False
        elif all(isinstance(val, (int, float, complex)) for val in self.amplitudes):
            return True
        else:
            raise ValueError('The amplitudes are NOT defined correctly.')

    # imaginary could be redefined as one of these properties

    def copy(self):
        return deepcopy(self)

    def addKet(self, ket, amplitude=None):
        if isinstance(ket, str):
            ket = tuple((ii, int(dim)) for ii, dim in enumerate(ket))
        elif isinstance(ket, (tuple, list)):
            pass  # seems legit
        else:
            raise ValueError('Invalid ket.')
        if ket in self.kets():
            print('The ket is going to be redefined.')
        else:  # the ket is included
            self.state[ket] = True
        if amplitude == None:
            self[ket] = defaultValues(1, self.imaginary)[0]
        else:
            self[ket] = amplitude

    def toCartesian(self):
        if self.imaginary == 'cartesian':
            print('The amplitudes are already in cartesian coordinates.')
            return None
        if self.is_weighted:
            if self.imaginary == False:
                for kk, vv in self.state.items():
                    self.state[kk] = (vv, 0)
            elif self.imaginary == 'polar':
                for kk, vv in self.state.items():
                    self.state[kk] = vv[0] * np.exp(1j * vv[1])
            else:
                raise ValueError('The propery `imaginary` is NOT defined correctly.')
        else:
            for kk, vv in self.state.items():
                self.state[kk] = True
        self.imaginary = 'cartesian'

    def toPolar(self):
        if self.imaginary == 'polar':
            print('The amplitudes are already in polar coordinates.')
            return None
        if self.is_weighted:
            if self.imaginary == False:
                for kk, vv in self.state.items():
                    self.state[kk] = (vv, 0)
            elif self.imaginary == 'cartesian':
                for kk, vv in self.state.items():
                    self.state[kk] = (abs(vv), np.angle(vv))
            else:
                raise ValueError('The propery `imaginary` is NOT defined correctly.')
        else:
            for kk, vv in self.state.items():
                self.state[kk] = (True, False)
        self.imaginary = 'polar'

    def rescale(self, constant):
        conversion = self.imaginary == 'polar'
        if conversion:
            self.toCartesian()
        for ket in self.state.keys():
            self.state[ket] *= constant
        if conversion:
            self.toPolar()

    def clamp(self, maximum=1, minimum=None):  # , rescale=False):
        if maximum >= 0:
            if minimum == None:
                minimum = - maximum
            for ket, amplitude in self.graph.items():
                self.state[ket] = max(minimum, min(amplitude, maximum))
        else:
            raise ValueError('Introduce a positive maximum.')

    def normalize(self):
        self.rescale(constant=1 / self.norm)

    def targetEquation(self, state_catalog=None, imaginary=None):
        if imaginary is None:
            return th.targetEquation(self.kets, amplitudes=self.amplitudes, state_catalog=state_catalog,
                                     imaginary=self.imaginary)
        else:
            return th.targetEquation(self.kets, amplitudes=self.amplitudes, state_catalog=state_catalog,
                                     imaginary=imaginary)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 12 11:54:12 2022

@author: janpe
"""

import pytheus.theseus as th
from collections import Counter
import itertools
import numpy as np
from itertools import combinations
from itertools import combinations_with_replacement
from pytheus.custom_loss.assembly_index import assembly_index, top_n_assembly
import logging

log = logging.getLogger(__name__)


def brutal_covers(cnfg, graph):
    # this functions is sometimes used instead of findEdgecovers, should be much slower in general, but faster at
    # applying topological constraints atm

    non_output_verts = [vert for vert in cnfg["verts"] if vert not in cnfg["out_nodes"]]
    # minimum number of edges that can cover non output vertices
    min_edges = len(non_output_verts) // 2
    # maximum number of edges, perfect matching order of whole graph
    max_edges = len(cnfg["verts"]) // 2
    orders = list(range(min_edges, max_edges + 1))

    avail_colors = th.edgeBleach(graph.edges)
    tmp_raw_covers = []
    for num_edges in orders:
        tmp_raw_covers += combinations_with_replacement(list(avail_colors.keys()), num_edges)

    raw_covers = []
    ii = 0
    for cover in tmp_raw_covers:
        ii += 1
        keep = True
        counter = Counter(list(itertools.chain(*cover)))

        # check inputs and single emitters
        for vert in cnfg["in_nodes"] + cnfg["single_emitters"]:
            # if any are covered more than once, don't keep edge cover
            if counter[vert] != 1:
                keep = False

        if cnfg['number_resolving']:
            for vert in cnfg["anc_detectors"]:
                if counter[vert] != 1:
                    keep = False
        else:
            for vert in cnfg["anc_detectors"]:
                if counter[vert] == 0:
                    keep = False

        # if in maximum order, we select for events where the correct total number of photons go into output paths
        if len(cover) == max_edges:
            sum = 0
            for vert in cnfg["out_nodes"]:
                sum += counter[vert]
            if sum != len(cnfg["out_nodes"]):
                keep = False

        # if all conditions are met, use edgecover for norm
        if keep:
            raw_covers.append(cover)
    painted_covers = []
    for cover in raw_covers:
        for coloring in itertools.product(*[avail_colors[edge] for edge in cover]):
            color_cover = [edge + color for edge, color in zip(cover, coloring)]
            painted_covers.append(sorted(color_cover))
    return [[tuple(ed) for ed in graph] for graph in np.unique(painted_covers, axis=0)]


def heralded_covers(cnfg, graph):
    # calls findEdgecovers and applies topological constraints

    non_output_verts = [vert for vert in cnfg["verts"] if vert not in cnfg["out_nodes"]]
    # minimum number of edges that can cover non output vertices
    min_edges = len(non_output_verts) // 2
    # maximum number of edges, perfect matching order of whole graph
    max_edges = len(cnfg["verts"]) // 2
    orders = list(range(min_edges, max_edges + 1))
    try:
        if cnfg["novac"]:
            orders = [max_edges]
    except KeyError:
        pass
    # find edge suitable edge covers for all possible numbers of edges
    tmp_edgecovers = []
    for num_edges in orders:
        tmp_edgecovers += th.findEdgeCovers(graph.edges, nodes_left=non_output_verts, edges_left=num_edges)
    # select for edgecovers that fulfill conditions
    edgecovers = []
    for cover in tmp_edgecovers:
        keep = True
        counter = Counter(list(itertools.chain(*th.edgeBleach(cover).keys())))

        # check inputs and single emitters
        for vert in cnfg["in_nodes"] + cnfg["single_emitters"]:
            # if any are covered more than once, don't keep edge cover
            if counter[vert] != 1:
                keep = False

        if cnfg['number_resolving']:
            for vert in cnfg["anc_detectors"]:
                if counter[vert] != 1:
                    keep = False
        else:
            for vert in cnfg["anc_detectors"]:
                if counter[vert] == 0:
                    keep = False

        # if in maximum order, we select for events where the correct total number of photons go into output paths
        if len(cover) == max_edges:
            sum = 0
            for vert in cnfg["out_nodes"]:
                sum += counter[vert]
            if sum != len(cnfg["out_nodes"]):
                keep = False

        # if all conditions are met, use edgecover for norm
        if keep:
            edgecovers.append(cover)
    return edgecovers


def count_rate(graph, target_state, cnfg):
    # can be used for state-creation/measurements/gates, post-selected/heralded

    # set up target equation
    target = target_state.targetEquation(state_catalog=graph.state_catalog, imaginary=cnfg["imaginary"])
    # get variable names
    variables = th.stringEdges(graph.edges, imaginary=cnfg["imaginary"])

    # non-heralded, post-selection case
    if not cnfg["heralding_out"]:
        # only looking at perfect matchings
        graph.getNorm()
        norm = graph.norm
    # heralded case, more complicated selection rules
    else:
        if not cnfg["brutal_covers"]:
            edgecovers = heralded_covers(cnfg, graph)
        else:
            edgecovers = brutal_covers(cnfg, graph)
        cat = th.stateCatalog(edgecovers)
        norm = th.writeNorm(cat, imaginary=cnfg["imaginary"])
    lambdaloss = "".join(["1-", target, "/(1+", norm, ")"])
    func, lossstring = th.buildLossString(lambdaloss, variables)
    return func


def fidelity(graph, target_state, cnfg):
    # can be used for state-creation/measurements/gates, post-selected/heralded

    # set up target equation
    target = target_state.targetEquation(state_catalog=graph.state_catalog, imaginary=cnfg["imaginary"])
    # get variable names
    variables = th.stringEdges(graph.edges, imaginary=cnfg["imaginary"])

    # non-heralded, post-selection case
    if not cnfg["heralding_out"]:
        # only looking at perfect matchings
        graph.getNorm()
        norm = graph.norm
    # heralded case, more complicated selection rules
    else:
        if not cnfg["brutal_covers"]:
            edgecovers = heralded_covers(cnfg, graph)
        else:
            edgecovers = brutal_covers(cnfg, graph)
        cat = th.stateCatalog(edgecovers)
        norm = th.writeNorm(cat, imaginary=cnfg["imaginary"])
    lambdaloss = "".join(["1-", target, "/(0+", norm, ")"])
    func, lossstring = th.buildLossString(lambdaloss, variables)
    return func


"""
Number states in Fock basis
"""
def fock_fidelity(graph, target_state, num_anc, amplitudes, imaginary=False): #num_anc_pre

    # original ket is in the form: [((0,m),(1,n),(2,k1),(3,k2)...]; 
    # here the m, n, k1, k2 are the number of particles instead of dimensions
    
    # make the ket in the correct form 
    # [((0,0),(0,0)...,(1,0),(1,0),... (2,0),(2,0),...,(3,0),(3,0),...]
    # m times (0,0); n times (1,0); k1 times (2,0); k2 times (3,0), etc.
    # target_kets_temp = target_state.kets 
    # target_kets=[]
    # accum = [sum([[(i,0)]*max(0,n) for i,n in k ],[]) for k in target_kets_temp]
    # for ii in accum:
    #     target_kets.append(tuple(ii))   
        
         
    #all_verts= range(graph.num_nodes)#np.unique(list(itertools.chain(*th.edgeBleach(graph.edges).keys())))
    #anc_each_num=[int(i) for i in num_anc_pre]
    #state_each_num=[int(i) for i in target_state_str[0]] # all terms have the same photon number
    
    total_particle_num = len(target_state.kets[0])
    #total_particle_num=sum(state_each_num)+num_anc #sum(anc_each_num)
    
    # anc_position=all_verts[len(all_verts)-num_anc:]
    # anc_kets=[]
    # for jj in anc_position:
    #     anc_kets.append((jj,0))
        
    anc_nodes = list(range(graph.num_nodes - num_anc, graph.num_nodes))
    
    
    edgecover_target = list(itertools.combinations_with_replacement(graph.edges,int(total_particle_num/2)))
    cat = th.stateCatalog(edgecover_target)
    
    if len(anc_nodes)>0:
        for ket in list(cat.keys()):
            shopping = Counter(ket)
            if all(shopping[(ii,0)]==1 for ii in anc_nodes):
                pass
            else:
                del cat[ket]
          
    # this cause a problem in the optimizer as it actually does not use len(useful_deges)
    #useful_edges = sorted(set(sum(sum(cat.values(),[]), ())))
    #variables = th.stringEdges(useful_edges)
    
    variables = th.stringEdges(graph.edges, imaginary=imaginary)
    
    target = th.targetEquation(target_state.kets,amplitudes=amplitudes,state_catalog=cat)
    norm = th.writeNorm(cat)

   # epsilon = 1e-10
    epsilon = 0
    lambdaloss = f'1-({target})/({epsilon} + {norm})'   
   #  lambdaloss="".join(["1-", target, "/(0+", norm, ")"])
    
    func, lossstring = th.buildLossString(lambdaloss, variables)
    return func

def fock_countrate(graph, target_state, num_anc, amplitudes, imaginary=False): #num_anc_pre

  
    
    total_particle_num = len(target_state.kets[0])
        
    anc_nodes = list(range(graph.num_nodes - num_anc, graph.num_nodes))
    
    
    edgecover_target = list(itertools.combinations_with_replacement(graph.edges,int(total_particle_num/2)))
    cat = th.stateCatalog(edgecover_target)
    
    if len(anc_nodes)>0:
        for ket in list(cat.keys()):
            shopping = Counter(ket)
            if all(shopping[(ii,0)]==1 for ii in anc_nodes):
                pass
            else:
                del cat[ket]
          
    # this cause a problem in the optimizer as it actually does not use len(useful_deges)
    #useful_edges = sorted(set(sum(sum(cat.values(),[]), ())))
    #variables = th.stringEdges(useful_edges)
    
    variables = th.stringEdges(graph.edges, imaginary=imaginary)
    
    target = th.targetEquation(target_state.kets,amplitudes=amplitudes,state_catalog=cat)
    norm = th.writeNorm(cat)

   # epsilon = 1e-10
    epsilon = 1
    lambdaloss = f'1-({target})/({epsilon} + {norm})'   
   #  lambdaloss="".join(["1-", target, "/(1+", norm, ")"])
    
    func, lossstring = th.buildLossString(lambdaloss, variables)
    return func



def make_lossString_entanglement(graph, sys_dict: dict, imaginary=False,
                                 var_factor=0):
    """
    get the loss functions of a graph for the concurrence:
        C( |Psi> ) = √( 2 * ( 1 - TR_M( <Psi|Psi> ) ) ) 
        where TR_M is partial trace (in subsystem M)
        and return is sum over all possible bi-partition

    Parameters
    ----------
    edge_list : list
        list of all edges 
    sys_dict : dict
        that stores essential information about the quantuum system (see hf.get_sysdict)

    Returns
    -------
    func : function as object
        loss function in concurrence as lambda object of current graph.
    lossstring : String
        loss function as string

    """

    cat = graph.state_catalog
    target = th.entanglement_fast(cat, sys_dict, var_factor)
    # norm = th.Norm.fromDictionary(cat, real=sys_dict['real'])
    variables = th.stringEdges(graph.edges, imaginary=imaginary)

    lambdaloss = "".join(["", target])
    func, lossstring = th.buildLossString(lambdaloss, variables)

    return func


def sum_of_weights(inputgraph, cnfg):
    # test function for an arbitrary loss function (cnfg["loss_func"] = "lff", cnfg["lff_name"] = "sum_of_weights")
    return sum(np.abs(inputgraph.weights))


def loss_from_function(graph, cnfg=[]):
    # takes a graph and returns a function of a parameter vector
    def func(vector):
        inputgraph = graph
        for ii, edge in enumerate(graph.edges):
            inputgraph[edge] = vector[ii]
        # get function (defined in this module)
        function = globals()[cnfg["lff_name"]]
        return function(inputgraph, cnfg)

    return func


# optimizer optimizes first function in list
# at each optimization step, all functions in list are checked if they satisfy the thresholds
loss_dic = {'ent': [make_lossString_entanglement],
            'fid': [fidelity, count_rate],
            'cr': [count_rate, fidelity],
            'lff': [loss_from_function],
            'fockfid': [fock_fidelity,fock_countrate],
            'fockcr': [fock_countrate,fock_fidelity]
           }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import exists
from pathlib import Path
import re

from setuptools import setup, find_packages

author = 'artificial-scientist-lab'
email = 'cruizgo@proton.me, soeren.arlt@mpl.mpg.de, mario.krenn@mpl.mpg.de'
description = 'Theseus, a highly-efficient inverse-design algorithm for quantum optical experiments'
dist_name = 'suehtyp'
package_name = 'pytheus'
year = '2022'
url = 'https://github.com/artificial-scientist-lab/Pytheus'


def get_version():
    content = open(Path(package_name) / '__init__.py').readlines()
    return "1.2.3"
    raise Exception('Cannot extract version string.')


setup(
    name=dist_name,
    author=author,
    author_email=email,
    url=url,
    version=get_version(),
    packages=find_packages(),
    package_dir={dist_name: package_name},
    include_package_data=True,
    license='MIT',
    description=description,
    long_description=open('README.md').read() if exists('README.md') else '',
    long_description_content_type="text/markdown",
    install_requires=[
        'sphinx', 'numpy', 'scipy', 'matplotlib', 'termcolor', 'Click'
    ],
    python_requires=">=3.6",
    classifiers=['Operating System :: OS Independent',
                 'Programming Language :: Python :: 3',
                 ],
    platforms=['ALL'],
    py_modules=[package_name],
    entry_points={
        'console_scripts': [
            'pytheus = pytheus.cli:cli',
        ],
    }
)



""" Module: cli.py

Theseus uses the Click library as a command line interface. Definition of click commands
and error handling should be contained in the cli.py module.

Conversely, to avoid tight coupling of code, all "business" logic, i.e. actions that shall be
performed when submitting a command, should be part of other modules, but NOT of cli.py.

In case of errors of any sort, the other modules should raise exceptions that cli.py
may handle as desired.

This decoupling makes sure that Theseus can not only be used as a command line application
but can also be imported as a package in normal Python code.
"""

import os
import sys
import json

import click
import pkg_resources

import pytheus
from pytheus.main import run_main
from pytheus.analyzer import get_analyse
from pytheus.graphplot import plotFromFile


@click.group()
def cli():
    pass


@cli.command()
@click.argument('filename')
@click.option('--example', is_flag=True, default=False, help='Load input file from examples directory.')
def run(filename, example):
    """Run an input file."""
    try:
        # run main routine (read config, set up target, run optimization/store results)
        run_main(filename, example)
    except IOError as e:
        click.echo('ERROR:' + str(e))
        sys.exit(1)


@cli.command()
@click.argument('filename')
@click.option('--pdf', default="", help='save output to pdf')
def plot(filename, pdf):
    """Plot a solution file."""
    try:
        plotFromFile(filename, outfile=pdf)
    except IOError as e:
        click.echo('ERROR:' + str(e))
        sys.exit(1)


@cli.command()
@click.option('-d', '--which-directory', default=None,
              help='choose folder to analyze')
@click.option('-one', '--all-weights-plus-minus-one', is_flag=True,
              show_default=True,
              help='map all weights to plus minus one')
@click.option('-pm', '--create-perfect-machting-pdf', is_flag=True,
              show_default=True,
              help='bool if create pdf with all pms')
@click.option('-i', '--which-infos', default=['norm', 'ent', 'k'],
              multiple=True,
              show_default=True,
              help='list of which infos appear in info plot')
def analyze(which_directory, all_weights_plus_minus_one,
            create_perfect_machting_pdf, which_infos):
    """Run anlyzer tool depending on inputs."""
    try:
        get_analyse(which_directory,
                    all_weights_plus_minus_one=all_weights_plus_minus_one,
                    create_perfect_machting_pdf=create_perfect_machting_pdf,
                    which_infos=which_infos)
    except IOError as e:
        click.echo('ERROR:' + str(e))
        sys.exit(1)


@cli.command()
def list():
    """List all included examples."""
    configs_dir = pkg_resources.resource_filename(pytheus.__name__, 'graphs')
    files = sorted(os.listdir(configs_dir))

    walk = os.walk(configs_dir)
    for root, dirs, files in walk:
        for file in files:
            if file.startswith('config'):
                click.echo(
                print(root.split('/')[-1]))



import itertools
import numpy as np


def flatten_lists(the_lists):
    """
    takes a list as argument return flatten one

    """
    result = []
    for _list in the_lists:
        result += _list
    return result


def stateToString(state, ket=False):
    '''
    for readability, turn state array into string
    '''
    termlist = np.array(state)[:, :, 1]
    termstringlist = []
    for term in termlist:
        termstringlist.append(''.join([str(item) for item in term]))
    if ket:
        termstringlist = ['|' + term + '>' for term in termstringlist]
    return '+'.join(termstringlist)


def readableState(state):
    readable_dict = {}
    for key in state.kets:
        readable_dict[stateToString([key], ket=True)] = state[key]
    return readable_dict


def stringToTerm(termstring):
    '''
    used by makeState. turn ket given by string into tuples representing creator operators.
    example: 
        input: "0210"
        output: ((0, 0), (1, 2), (2, 1), (3, 0))
    '''
    return tuple([tuple([i, int(col)]) for i, col in enumerate(termstring)])


def makeState(statestring: str) -> list:
    '''
    turn state given as string into state to be used by pytheus
    example:
        input: "0000+1111+2222"
        output: [((0, 0), (1, 0), (2, 0), (3, 0)),
                 ((0, 1), (1, 1), (2, 1), (3, 1)),
                 ((0, 2), (1, 2), (2, 2), (3, 2))]
    '''
    terms = statestring.split('+')
    return [stringToTerm(filter(str.isdigit, term)) for term in terms]


def makeUnicolor(edge_list, num_nodes):
    '''
    simplify edge list by deleting all multicolor edges between data nodes
    '''
    return [edge for edge in edge_list if
            (((edge[0] not in range(num_nodes)) or (edge[1] not in range(num_nodes))) or (edge[2] == edge[3]))]


def removeConnections(edge_list, connection_list):
    '''
    removes all edges that connect certain pairs of vertices.

    example:
    input: edge_list, [[0,1],[3,5]]
    output: edge_list without any edges that connect 0-1 or 3-5.
    '''
    new_edge_list = edge_list
    for connection in connection_list:
        new_edge_list = [edge for edge in new_edge_list if (edge[0] != connection[0] or edge[1] != connection[1])]
    return new_edge_list


def prepEdgeList(edge_list, cnfg):
    """
    Restrict starting graph as given by config.
    """
    try:
        if cnfg['unicolor']:
            num_data_nodes = len(cnfg['target_state'][0])
            edge_list = makeUnicolor(edge_list, num_data_nodes)
    except KeyError:
        pass

    removed_connections = []
    disjoint_nodes = []
    try:
        disjoint_nodes += cnfg['in_nodes']
    except KeyError:
        pass
    try:
        disjoint_nodes += cnfg['single_emitters']
    except KeyError:
        pass
    removed_connections += list(itertools.combinations(disjoint_nodes,2))
    try:
         removed_connections += cnfg['removed_connections']
    except KeyError:
        pass
    edge_list = removeConnections(edge_list,removed_connections)
    return edge_list


def get_all_bi_partions(num_par: int, lenght=None):

    """
    returns all bi-partions as a generator for a given number of particles:

        e.g. : num_par = 3 : [([0], [1, 2]), ([1], [0, 2]), ([2], [0, 1])]

    """
    if lenght is None or lenght == 'all':
        def check_len(bipar):
            return True
    else:
        assert (type(lenght) is int and int(num_par / 2) >= lenght), \
            "invalid lenght given(or Typeerror): int(num_par/2) > given lenght"

        def check_len(bipar):
            return len(bipar) == lenght

    S = {i for i in range(num_par)}
    doubles = []
    for ll in range(1, int(len(S) / 2) + 1):
        combinations = set(itertools.combinations(S, ll))
        for oneC in combinations:
            if sorted(list(oneC)) not in doubles:
                bipar = sorted(list(oneC))
                if check_len(bipar):
                    yield (bipar, sorted(list(S - set(oneC))))
            doubles.append(sorted(list(S - set(oneC))))


def get_all_kets_for_given_dim(dimension: list, type_return=int):
    """
    get all possible kets for given dimension  (1=ancilla)
    e.g: input = [2,2,1,1] -> [0, 1, 10, 11]  for int 
                       -> ['00', '01', '10', '11'] for str
    """
    if dimension.count(1) != 0:
        dimension = dimension[:-dimension.count(1)]
    if type_return == int:
        return list([int("".join(map(str, x))) for x in
                     itertools.product(*[[t for t in range(i)]
                                         for i in dimension])])
    if type_return == str:
        return list(["".join(map(str, x)) for x in itertools.product(
            *[[t for t in range(i)] for i in dimension])])


def get_complement(ls: tuple, whole_list: list):
    return tuple(set([x for x in whole_list]) - set(ls))


def get_sysdict(dimensions_of_H: list, bipar_for_opti='all', imaginary = False):
    """
    a dict to store
        - number of particles: sysdict['num_particles']
        - all possible bipartions: sysdict['all_biparations']
        - all dimension for each particle: sysdict['dimensions']
        - how many solution should be produced: sysdict["samples"]
    """
    sysdict = dict()
    sysdict['dimensions'] = dimensions_of_H
    sysdict['num_ancillas'] = dimensions_of_H.count(1)
    sysdict['num_particles'] = len(dimensions_of_H) - sysdict['num_ancillas']
    sysdict['all_states'] = [tuple([(idx, int(ket[idx])) for idx in
                                    range(sysdict['num_particles'])])
                             for ket in
                             get_all_kets_for_given_dim(dimensions_of_H, str)]

    sysdict['dim_total'] = np.product(dimensions_of_H)
    sysdict['bipar_for_opti'] = list(
        get_all_bi_partions(sysdict['num_particles'], bipar_for_opti))
    sysdict['imaginary'] = imaginary
    return sysdict
